#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from ssht00ls.classes.config import *
from ssht00ls.classes.aliases import aliases
from ssht00ls.classes.agent import agent
import ssht00ls.classes.ssync.utils as ssync_utils 
from ssht00ls.classes.ssh import ssh

# sync daemon cache.
def sync():
	dir = Files.Directory(path=f"{cache.path}/daemons/")
	if not dir.fp.exists(): dir.create()
	for path in dir.paths(recursive=False):
		removed = False
		fp = FilePath(gfp.name(path=path).replace("\\","/"))
		if not fp.exists(): 
			os.system(f"rm -fr {path}")
			removed = True
		if not removed:
			status = str(cache.get(id=fp.path, group="daemons"))
			if "*running*" in status and not running(fp.path): 
				os.system(f"rm -fr {path}")
				removed = True

# check if daemon is running.
def running(path):
	
	# check status.
	cache_path = gfp.absolute(gfp.clean(path.split(" (d)")[0], remove_last_slash=True))
	status = str(cache.get(id=cache_path, group="daemons"))
	if not "*running*" in status: return False
	
	# check stamp.
	if "timestamp=" in status:
		timestamp = status.split("(timestamp=")[1].split(")")[0]
		date = Date()
		decreased = date.decrease(date.seconds_timestamp, seconds=SSYNC_DAEMON_SLEEPTIME*5, format=date.seconds_timestamp_format)
		if date.compare(current=decreased, comparison=timestamp, format=date.seconds_timestamp_format) in ["past", "present"]:
			return False
		else:
			return True
	# should not happen.
	else:
		raise ValueError(f"Should not happen, *running* is in status but no timestamp present, (status: {status}).")

# list daemons.
def status(path=None):
	if path == None:
		dict = {}
		for i in list():
			dict[i] = status(path=i)
		return dict
	else:
		sync()
		cache_path = gfp.absolute(gfp.clean(path.split(" (d)")[0], remove_last_slash=True))
		return cache.get(id=cache_path, group="daemons")

# list daemons.
def list(filter=None):
	sync()
	dir = Files.Directory(path=f"{cache.path}/daemons/")
	listed = []
	for path in dir.paths(recursive=False):
		fp = FilePath(gfp.name(path=path).replace("\\","/"))
		listed.append(fp.path)
	return listed

# stop a daemon.
def stop(path, timeout=SSYNC_DAEMON_SLEEPTIME*10, sleeptime=1):
	if timeout <= 10: timeout = 10
	cache_path = gfp.absolute(gfp.clean(path.split(" (d)")[0], remove_last_slash=True))
	cache.set(id=cache_path, data="*stop*", group="daemons")
	stopped = False
	for i in range(int(timeout/sleeptime)):
		status_ = str(cache.get(id=cache_path, group="daemons"))
		if "*stopped*" in status_ or "*crashed*" in status_:
			stopped = True
			break
		time.sleep(sleeptime)
	if stopped:
		return dev0s.response.success(f"Successfully stopped ssht00ls daemon [{path}].")
	else:
		return dev0s.response.error(f"Failed to stop ssht00ls daemon [{path}].")

# the daemon object class.
class Daemon(Thread):
	def __init__(self, attributes={}):

		# defaults.
		Traceback.__init__(self, traceback="ssht00ls.ssync.daemons.Daemon", raw_traceback="ssht00ls.classes.ssync.daemons.Daemon")

		# attributes.
		self.utils = ssync_utils
		Thread.__init__(self)
		self.assign(attributes)
		self.path = gfp.clean(self.path)
		self.id = f"{self.alias}:{self.remote} {self.path}"
		try:self.log_level
		except: self.log_level = -1
		self.last_index = {}
		self.last_remote_index = {}
		self.cache_path = gfp.absolute(gfp.clean(self.path.split(" (d)")[0], remove_last_slash=True))

		#
	def subpath(self, fullpath, remote=False, append_slash=False):
		if remote: path = self.remote
		else: path = self.path
		s = ""
		if append_slash: s = "/"
		return fullpath.replace(path+s, "").replace("//","/").replace("//","/").replace("//","/").replace("//","/").replace("//","/").replace("//","/").replace("//","/").replace("//","/").replace("//","/").replace("//","/")
	def fullpath(self, subpath, remote=False, append_slash=False):
		if remote: path = self.remote
		else: path = self.path
		s = ""
		if append_slash: s = "/"
		return f"{path}/{subpath}{s}".replace("//","/").replace("//","/").replace("//","/").replace("//","/").replace("//","/").replace("//","/").replace("//","/").replace("//","/").replace("//","/").replace("//","/").replace("//","/").replace("//","/").replace("//","/").replace("//","/").replace("//","/")
	def run(self):

		# logs.
		if self.log_level >= 0: 
			loader = dev0s.console.Loader(f"Checking daemon {self.id}", interactive=dev0s.defaults.options.interactive)

		# checks.
		status = str(cache.get(id=self.cache_path, group="daemons"))
		if not running(self.cache_path):
			if self.log_level >= 0: loader.stop(success=False)
			self.crash(f"ssht00ls daemon ({self.id}): Path [{self.path}] is not running (status: {status}).")
		if self.log_level >= 0: 
			loader.mark(f"Checking mountpoint {self.path}")
		if not Files.exists(self.path):
			if self.log_level >= 0: loader.stop(success=False)
			self.crash(f"ssht00ls daemon ({self.id}): Path [{self.path}] does not exist.")
		if not os.path.isdir(self.path):
			if self.log_level >= 0: loader.stop(success=False)
			self.crash(f"ssht00ls daemon ({self.id}): Path [{self.path}] is not a directory.")

		# check alias.
		if self.log_level >= 0: 
			loader.mark(f"Checking alias {self.alias}")
		response = aliases.check(self.alias)
		if not response["success"]: 
			if self.log_level >= 0: loader.stop(success=False)
			self.crash(response=response)

		# get index.
		if self.log_level >= 0: 
			loader.mark(f"Indexing [{self.path}] & [{self.alias}:{self.remote}].")
		response = self.index(short=True)
		if not response["success"]: 
			if self.log_level >= 0: loader.stop(success=False)
			self.crash(response=response)

		# start success.
		if self.log_level >= 0: 
			loader.mark(f"Starting daemon {self.id}")
			loader.stop()

		# start.
		timestamp_count = 0
		while True:

			# check stop command.
			status = str(cache.get(id=self.cache_path, group="daemons"))
			if "*stop*" in status or "*unmounting*" in status:
				break

			# if file no longer exists stop daemon.
			if status == "none" or not Files.exists(self.path): 
				self.crash(f"ssht00ls daemon ({self.id}): Mounted directory [{self.path}] from [{self.alias}:{self.remote}] no longer exists.")

			# sync.
			response = self.sync(attempts=3, delay=[3, 5, 10])
			if self.log_level >= 1:
				dev0s.response.log(response=response)
			if not response["success"]: 
				self.crash(f"ssht00ls daemon ({self.id}) encountered an error while synchronizing: {response.error}", unmount=False)

			# set timestamp.
			do = False
			if self.sleeptime >= 1: do = True
			elif timestamp_count >= 1 / self.sleeptime:
				do = True
				timestamp_count = 0
			timestamp_count += 1
			if do:
				status = str(cache.get(id=self.cache_path, group="daemons"))
				if "*running*" in status:
					cache.set(id=self.cache_path, group="daemons", data=f"*running* (timestamp={Date().seconds_timestamp})")

			# sleep.
			time.sleep(self.sleeptime)

		# stop.
		self.stop()

		#
	def stop():
		response = self.unmount()
		if not response["success"]: self.crash(f"ssht00ls daemon ({self.id}) error: {response['error']}")
		cache.set(id=self.cache_path, data="*stopped*", group="daemons")
		if self.log_level >= 0: print(f"Stopped daemon {self.id}")
	def crash(self, error=None, response=None, unmount=True):
		if response != None: error = response["error"]
		if unmount:
			response = self.unmount()
			if not response["success"]: 
				cache.set(id=self.cache_path, data="*crashed*", group="daemons")
				raise ValueError(f"ssht00ls daemon ({self.id}) error: {response['error']}")
		cache.set(id=self.cache_path, data="*crashed*", group="daemons")
		raise ValueError(error)
	def unmount(self):
		if self.mode == "mount":
			if Files.exists(self.path):
				response = self.sync(attempts=5, delay=[1, 3, 5, 10, 15])
				if not response["success"]: return response
				cache.set(id=self.cache_path, data="*unmounting*", group="daemons")
				time.sleep(0.5)
				response = self.delete(self.path, remote=False, subpath=False,)
				if not response["success"]: return response
		return dev0s.response.success("Successfully unmounted.")
	def delete(self, path, remote=False, subpath=True):

		# sandbox.
		if isinstance(path, str):
			path = self.utils.serialize_path(gfp.clean(path))
			if subpath: path = self.fullpath(path, remote=remote)
			str_id = path
		else:
			new = []
			for i in path: 
				i = self.utils.serialize_path(gfp.clean(i))
				if subpath: i = self.fullpath(i, remote=remote)
				new.append(i)
			path = new
			str_id = f"{len(path)} path(s)"
		if self.sandbox: 
			if remote:
				msg = f"Sandbox enabled, skip deletion of {self.alias}:{str_id}."
				if self.log_level >= 0: print(msg)
				return dev0s.response.success(msg)
			else:
				msg = f"Sandbox enabled, skip deletion of {str_id}."
				if self.log_level >= 0: print(msg)
				return dev0s.response.success(msg)

		# logs.
		if self.log_level >= 1:
			if remote: print(f"Deleting {self.alias}:{str_id}.")
			else:print(f"Deleting {str_id}.")
		
		# local to remote.
		if remote:
			if isinstance(path, str):
				cmd = f"""ssh {DEFAULT_SSH_OPTIONS} {self.alias} " printf 'y' | rm -fr '{path}' " """
			else:
				c = 0
				cmd = f"""ssh {DEFAULT_SSH_OPTIONS} {self.alias} " """
				for i in path:
					if c == 0:
						cmd += f"""printf 'y' | rm -fr '{i}'"""
					else:
						cmd += f""" && printf 'y' | rm -fr '{i}'"""
					c += 1
				cmd += ' " '
			response = self.utils.execute(command=cmd,)
			if not response["success"]: return response
			#response = ssh.utils.test_path(alias=self.alias, path=path)
			#if response.error != None and f"{path} does not exist" not in response.error:
			#	return dev0s.response.error(f"Failed to delete {self.alias}:{path}, error: {response.error}")
			if response.output != "":
				return dev0s.response.error(f"Failed to delete {self.alias}:{str_id}, error: {output}")
			return dev0s.response.success(f"Successfully deleted {self.alias}:{str_id}")

		# remote to local.
		else:
			if isinstance(path, str):
				cmd = f"""printf 'y' | rm -fr '{path}' """
			else:
				cmd, c = "", 0
				for i in path:
					if c  == 0:
						cmd += f"""printf 'y' | rm -fr '{i}'"""
					else:
						cmd += f""" && printf 'y' | rm -fr '{i}'"""
					c += 1
			response = self.utils.execute(command=cmd)
			if not response["success"]: return response
			if response.output != "":
				return dev0s.response.error(f"Failed to delete {str_id}, error: {output}")
			return dev0s.response.success(f"Successfully deleted {str_id}")
	def index(self, short=False):

		# index.
		remote_response = self.ssync.index(path=self.remote, alias=self.alias, checks=False, log_level=self.log_level)
		if not remote_response["success"]:  return remote_response
		response = self.ssync.index(path=self.path, log_level=self.log_level)
		if not response["success"]:  return response

		# clean indexes.
		index, remote_index, clean_index, remote_clean_index, all_paths = response["index"], remote_response["index"], {}, {}, []
		for path, mtime in index.items():
			subpath = self.subpath(path)
			if subpath not in all_paths: 
				all_paths.append(subpath)
			subpath = subpath.split(" (d)")[0] # unpack after appended to all_paths.
			try:
				clean_index[subpath] = index[path]
			except KeyError:
				return dev0s.response.error(f"Should not happen(#278363). (subpath: {subpath}), (path: {path}), (index: {index}), (remote index: {remote_index}).")
		for path, mtime in remote_index.items():
			subpath = self.subpath(path)
			if subpath not in all_paths: 
				all_paths.append(subpath)
			subpath = subpath.split(" (d)")[0] # unpack after appended to all_paths.
			try:
				remote_clean_index[subpath] = remote_index[path]
			except KeyError:
				return dev0s.response.error(f"Should not happen(#739639). (subpath: {subpath}), (path: {path}), (index: {index}), (remote index: {remote_index}).")

		# mismatches.
		synchronized, mismatches = self.mismatches(clean_index, remote_clean_index, all_paths)

		# short mode.
		if short in ["automatic", "auto"]:
			if not synchronized: short = False
		if short:
			if self.last_index == {}:
				self.last_index["0"] = clean_index
			if self.last_remote_index == {}:
				self.last_remote_index["0"] = remote_clean_index
			return dev0s.response.success(f"Successfully indexed [{self.alias}:{self.path}] & [{self.path}].", {
				"synchronized":synchronized,
				"index":index,
				"remote_index":remote_index,
				"mismatches":mismatches,
			})

		# process mismatches.
		else: return self.process_mismatches(clean_index, remote_clean_index, mismatches)

		#
	def mismatches(self, clean_index, remote_clean_index, all_paths):
		# current version.
		mismatches = {}
		for fullpath in all_paths:
			subpath = fullpath.split(" (d)")[0]
			lmtime, rmtime, synchronized = None, None, True
			try: 
				lmtime = clean_index[subpath]
			except KeyError: a=1
			try:
				rmtime = remote_clean_index[subpath]
			except KeyError: a=1
			if lmtime != None or rmtime != None:
				synchronized = lmtime == rmtime
			# skip dirs with not the same mtime but that do both have a mtime.
			if " (d)" in fullpath and lmtime != None and rmtime != None and lmtime != rmtime:
				if self.log_level >= 3:
					print(f"Revert directory {subpath} from synchronized {synchronized} to True")
				synchronized = True
			if not synchronized: 
				mismatches[subpath] = {
					"path":subpath,
					"directory":" (d)" in fullpath,
					"empty_directory":" (d)" in fullpath and " (e)" in fullpath,
					"local_mtime":lmtime,
					"remote_mtime":rmtime,
				}
		synchronized = len(mismatches) == 0
		if self.log_level >= 3: print(f'synchronized: {synchronized} clean_index: {clean_index}, remote_clean_index: {remote_clean_index} mismatches: {mismatches}.')
		return synchronized, mismatches
		"""
		mismatches, likeys, rikeys = {}, list(index.keys()), list(remote_index.keys())
		for i in likeys:
			if i not in rikeys:
				try: lmtime = clean_index[i]
				except: lmtime = None
				try: rmtime = remote_clean_index[i]
				except: rmtime = None
				if lmtime != rmtime:
					mismatches[i] = {
						"path":i,
						"local_mtime":lmtime,
						"remote_mtime":rmtime,
					}
		for i in rikeys:
			if i not in likeys:
				try: lmtime = clean_index[i]
				except: lmtime = None
				try: rmtime = remote_clean_index[i]
				except: rmtime = None
				if lmtime != rmtime:
					mismatches[i] = {
						"path":i,
						"local_mtime":lmtime,
						"remote_mtime":rmtime,
					}
		return dev0s.response.error(f"Failed to synchronize [{self.alias}:{self.remote}] & [{self.path}], mismatches: {mismatches}].")
		"""
	def process_mismatches(self, clean_index, remote_clean_index, mismatches):

		# iterate.
		updates, deletions = {}, {}
		for subpath, info in mismatches.items():

			# vars.
			directory = info["directory"]
			empty_directory = info["empty_directory"]
			subpath = subpath.split(" (d)")[0] # just to be sure.
			lfullpath, rfullpath = self.fullpath(subpath), self.fullpath(subpath, remote=True)
			try: lmtime = clean_index[subpath]
			except KeyError: lmtime = None
			try: rmtime = remote_clean_index[subpath]
			except KeyError: rmtime = None
			last_lmtime = self.get_last_index(subpath)
			last_rmtime = self.get_last_index(subpath, remote=True)

			# wanted vars.
			local_to_remote, remote_to_local = False, False
			options = []
			
			# should not happen.
			if rmtime == None and lmtime == None: 
				self.set_last_index(clean_index) ; self.set_last_index(remote_clean_index, remote=True)
				return dev0s.response.error(f"No remote & local modification time present. (path: {subpath}), (rmtime: {rmtime}), (lmtime: {lmtime}), (last_rmtime: {last_rmtime}), (last_lmtime: {last_lmtime}), index: {clean_index}, remote index: {remote_clean_index}.")

			# one missing.
			elif rmtime == None or lmtime == None:
				
				# remote deleted a file.
				if rmtime == None and lmtime != None and (last_lmtime != None or last_rmtime != None):
					if self.log_level >= 3:
						print(f"Remote deleted a file (path: {subpath}) (rmtime: {rmtime}), (lmtime: {lmtime}), (last_lmtime: {last_lmtime}).")
					remote_to_local = True
					options.append("delete")

				# local deleted a file.
				elif rmtime != None and lmtime == None and last_lmtime != None:
					if self.log_level >= 3:
						print(f"Local deleted a file (path: {subpath}) (rmtime: {rmtime}), (lmtime: {lmtime}), (last_lmtime: {last_lmtime}).")
					local_to_remote = True
					options.append("delete")

				# local created a file.
				elif rmtime == None and lmtime != None and last_lmtime == None:
					if self.log_level >= 3:
						print(f"Local created a file (path: {subpath}) (rmtime: {rmtime}), (lmtime: {lmtime}), (last_lmtime: {last_lmtime}).")
					local_to_remote = True

				# remote created a file.
				elif rmtime != None and lmtime == None and last_lmtime == None:
					if self.log_level >= 3:
						print(f"Remote created a file (path: {subpath}) (rmtime: {rmtime}), (lmtime: {lmtime}), (last_lmtime: {last_lmtime}).")
					remote_to_local = True

				# should not happen.
				else:
					self.set_last_index(clean_index) ; self.set_last_index(remote_clean_index, remote=True)
					return dev0s.response.error(f"Should not happen (#3243443). (path: {subpath}), (rmtime: {rmtime}), (lmtime: {lmtime}), (last_rmtime: {last_rmtime}), (last_lmtime: {last_lmtime}), index: {clean_index}, remote index: {remote_clean_index}.")

			# both present.
			elif rmtime != None and lmtime != None:

				# same mtime.
				if str(rmtime) == str(lmtime):
					a=1
				
				# # synchronize remote to local.
				elif rmtime > lmtime:
					remote_to_local = True

				# # synchronize local to remote.
				elif rmtime < lmtime:
					local_to_remote = True

				# should not happen.
				else:
					self.set_last_index(clean_index) ; self.set_last_index(remote_clean_index, remote=True)
					return dev0s.response.error(f"Unable to compare rmtime: {rmtime} & lmtime: {lmtime}. (path: {subpath}), (rmtime: {rmtime}), (lmtime: {lmtime}), (last_rmtime: {last_rmtime}), (last_lmtime: {last_lmtime}), index: {clean_index}, remote index: {remote_clean_index}.")

			# exceptions.
			else:
				self.set_last_index(clean_index) ; self.set_last_index(remote_clean_index, remote=True)
				return dev0s.response.error(f"Should not happen (#407294). (path: {subpath}), (rmtime: {rmtime}), (lmtime: {lmtime}), (last_rmtime: {last_rmtime}), (last_lmtime: {last_lmtime}), index: {clean_index}, remote index: {remote_clean_index}.")

			# do not remove this exception.
			# it is required for safe edits, to make sure a dir never gets synced unless it is created / removed.
			# also required by the multiprocessing idexing.
			if directory and local_to_remote != None and remote_to_local != None and "delete" not in options and not empty_directory:
				pass
			else:

				# add to updates.
				if local_to_remote and remote_to_local:
					self.set_last_index(clean_index) ; self.set_last_index(remote_clean_index, remote=True)
					return dev0s.response.error(f"Can not synchronize both remote to local & local to remote (rmtime: {rmtime}) (lmtime: {lmtime}), (last_rmtime: {last_rmtime}) (last_lmtime: {last_lmtime}), index: {clean_index}, remote index: {remote_clean_index}.")
				if "delete" in options: 
					if self.log_level >= 1: 
						if local_to_remote:
							print(f"Deletion required {lfullpath} {self.alias}:{rfullpath} (rmtime: {rmtime}) (lmtime: {lmtime}), (last_rmtime: {last_rmtime}) (last_lmtime: {last_lmtime})")
						else:
							print(f"Deletion required {self.alias}:{rfullpath} {lfullpath} (rmtime: {rmtime}) (lmtime: {lmtime}), (last_rmtime: {last_rmtime}) (last_lmtime: {last_lmtime})")
					deletions[subpath] = {
						"options":options,
						"remote_to_local":remote_to_local,
						"local_to_remote":local_to_remote,
						"directory":directory,
						"empty_directory":empty_directory,}
				elif remote_to_local or local_to_remote:
					if self.log_level >= 1: 
						if local_to_remote:
							print(f"Update required {lfullpath} {self.alias}:{rfullpath} (rmtime: {rmtime}) (lmtime: {lmtime}), (last_rmtime: {last_rmtime}) (last_lmtime: {last_lmtime})")
						else:
							print(f"Update required {self.alias}:{rfullpath} {lfullpath} (rmtime: {rmtime}) (lmtime: {lmtime}), (last_rmtime: {last_rmtime}) (last_lmtime: {last_lmtime})")
					updates[subpath] = {
						"options":options,
						"remote_to_local":remote_to_local,
						"local_to_remote":local_to_remote,
						"directory":directory,
						"empty_directory":empty_directory,}
		# handler.
		self.set_last_index(clean_index) ; self.set_last_index(remote_clean_index, remote=True)
		return dev0s.response.success(f"Successfully indexed [{self.alias}:{rfullpath}] & [{lfullpath}].", {
			"synchronized":len(updates) == 0 and len(deletions) == 0,
			"updates":updates,
			"deletions":deletions,
			"index":clean_index,
			"remote_index":remote_clean_index,
		})
	def local_to_remote(self, path, info, directory=False, empty_directory=False, forced=False, delete=False, command=False):
		lfullpath, rfullpath = self.fullpath(path, append_slash=directory), self.fullpath(path, remote=True, append_slash=directory)
		if self.log_level >= 1: print(f"Synchronizing {lfullpath} to {self.alias}:{rfullpath} (directory: {directory}) (delete: {delete}) (forced: {forced}).")
		return self.ssync.push(
			path=lfullpath, 
			alias=self.alias, 
			remote=rfullpath, 
			directory=directory,
			empty_directory=empty_directory,
			delete=delete,
			forced=forced,
			safe=False,
			accept_new_host_keys=True,
			checks=False,
			command=command,
			log_level=self.log_level,)
	def remote_to_local(self, path, info, directory=False, empty_directory=False, forced=False, delete=False, command=False):
		lfullpath, rfullpath = self.fullpath(path, append_slash=directory), self.fullpath(path, remote=True, append_slash=directory)
		if self.log_level >= 1: print(f"Synchronizing {self.alias}:{rfullpath} to {lfullpath} (directory: {directory}) (delete: {delete}) (forced: {forced}).")
		return self.ssync.pull(
			path=lfullpath, 
			alias=self.alias, 
			remote=rfullpath, 
			directory=directory,
			empty_directory=empty_directory,
			delete=delete,
			forced=forced,
			safe=False,
			accept_new_host_keys=True,
			checks=False,
			command=command,
			log_level=self.log_level,)
	def get_last_index(self, id, remote=False):
		if not remote: indexes = self.last_index
		else: indexes = self.last_remote_index
		value = None
		for _, index in indexes.items():
			try: value = index[id]
			except KeyError: a=1
			if value not in ["None", None, "none"]:
				break
		if self.log_level >= 6:
			print(f"Last index ({id}) {value}) (remote: {remote}), indexes: {indexes}.")
		return value
	def set_last_index(self, index, remote=False, depth=25):
		if not remote: indexes = self.last_index
		else: indexes = self.last_remote_index
		count = len(indexes)
		if count > depth:
			new = {}
			for key,value in indexes.items():
				key = int(key)
				if key > 0:
					new[str(key-1)] = value
			new[str(depth)] = index
			if not remote: self.last_index = new
			else: self.last_remote_index = new
		else:
			indexes[str(len(indexes))] = index
			if not remote: self.last_lmtime = indexes
			else: self.last_rmtime = indexes
	def reset_last_index(self, id, remote=False):
		if remote == "both":
			self.reset_last_index(id, remote=False)
			self.reset_last_index(id, remote=True)
		else:
			if not remote: indexes = self.last_index
			else: indexes = self.last_remote_index
			for depth in list(indexes.keys()):
				try: 
					index = indexes[depth]
					index[id]
					indexes[depth][id] = None
				except KeyError: a=1
			if not remote: self.last_index = indexes
			else: self.last_remote_index = indexes
	def sync(self, attempts=1, delay=3):
		
		# sync.
		response = self.__sync__()
		if not response["success"]:

			# delay.
			attempts -= 1
			if attempts > 0:
				if isinstance(delay, str):
					time.sleep(delay)
				else:
					time.sleep(delay[0])
					delay.pop(0)

				# recusrive.
				return self.sync(attempts=attempts, delay=delay)

				# handle network error.
				#if "Operation timed out" in response["error"]:
				#	return self.sync(attempts=attempts, delay=30)

				# default error.
				#else: return response


			# default error.
			else: return response

		# success.
		return response
		#
	def __sync__(self, multiprocessing=True, max_batch_size=150):
		
		# correct batch size.
		# to high produces ssh-agent errors.
		if max_batch_size >= 150:
			max_batch_size = 150

		# check synced.
		"""
		response = self.synchronized()
		if not response["success"]:  return response
		elif response.synchronized:
			if self.log_level > 0:
				print(f"Directories [{self.alias}:{self.path}] & [{self.path}] are already synchronized.")
			if self.log_level >= 3:
				return dev0s.response.success(f"Directories [{self.alias}:{self.path}] & [{self.path}] are already synchronized, index: {response.index}, remote index: {response.remote_index}.")
			else:
				return dev0s.response.success(f"Directories [{self.alias}:{self.path}] & [{self.path}] are already synchronized.")
		"""

		# get index.
		response = self.index(short="auto")
		if not response["success"]:  return response
		elif response.synchronized:
			if self.log_level >= 1:
				print(f"Directories [{self.alias}:{self.path}] & [{self.path}] are already synchronized.")
			if self.log_level >= 3:
				return dev0s.response.success(f"Directories [{self.alias}:{self.path}] & [{self.path}] are already synchronized, index: {response.index}, remote index: {response.remote_index}.")
			else:
				return dev0s.response.success(f"Directories [{self.alias}:{self.path}] & [{self.path}] are already synchronized.")
		updates,deletions = response.unpack(["updates", "deletions"])
		if multiprocessing and len(updates) < 50:
			if self.log_level >= 0:
				print(f"Disabling multiprocessing due to small batch size ({len(updates)}).")
			multiprocessing = False

		# parallel multiprocessing.
		if multiprocessing:
		
			# order updates by dir depth from deepest to min.
			# create depth index.
			dir_updates, depths, max_depth = {}, {}, 0
			for path, info in updates.items():
				if info["directory"]: dir_updates[path] = info
				else:
					depth = len(gfp.clean(path, remove_double_slash=True, remove_first_slash=True, remove_last_slash=True).split("/"))
					if depth >= max_depth: max_depth = depth
					try: depths[str(depth)]
					except KeyError: depths[str(depth)] = {}
					depths[str(depth)][path] = info
			
			# add batches from highest depth to lowest.
			update_batches, batch_size, inside_batch_size, last_depth = {}, 0, 0, None
			for i in range(max_depth+1):
				depth, found = max_depth-i, True
				try: depths[str(depth)]
				except KeyError: found = False
				if found:
					if (last_depth != None and last_depth != depth) or (inside_batch_size >= max_batch_size): 
						last_depth = depth
						batch_size += 1
						inside_batch_size = 0
					for path, info in depths[str(depth)].items():
						if inside_batch_size >= max_batch_size: 
							last_depth = depth
							batch_size += 1
							inside_batch_size = 0
						try: update_batches[str(batch_size)]
						except KeyError: update_batches[str(batch_size)] = {}
						update_batches[str(batch_size)][path] = info
						inside_batch_size += 1
					del depths[str(depth)]
			if depths != {}:
				return dev0s.response.error(f"Should not happen(#234938). (depths: {depths}), (updates: {updates}).")

			# add directories.
			for path, info in dir_updates.items():
				if inside_batch_size >= max_batch_size: 
					batch_size += 1
					inside_batch_size = 0
				try: update_batches[str(batch_size)]
				except KeyError: update_batches[str(batch_size)] = {}
				update_batches[str(batch_size)][path] = info
				inside_batch_size += 1

			# sum together as one command
			#print("UPDAT BACTHES:",update_batches)
			sum = True

			# iterate updates.
			commands = {}
			for batch_size, batch in update_batches.items():
				commands[batch_size] = []
				for path, info in batch.items():
					if self.log_level >= 1: print(f"Updating {path}.")

					# local to remote.
					if info["local_to_remote"]:
						if info["directory"] in [True, "True", "TRUE", "true"]: 
							commands[batch_size] += [self.local_to_remote(path, info, directory=True, command=sum)]
						else: 
							commands[batch_size] += [self.local_to_remote(path, info, command=sum)]

					# remote to local.
					elif info["remote_to_local"]:
						if info["directory"] in [True, "True", "TRUE", "true"]: 
							commands[batch_size] += [self.remote_to_local(path, info, directory=True, command=sum)]
						else: 
							commands[batch_size] += [self.remote_to_local(path, info, command=sum)]

			# execute commands.
			if sum:
				for batch_size, batch in commands.items():
					loader, c = None, len(batch)
					if self.log_level >= 0: loader = f"Synchronizing {c} file(s)."
					msg = f"Successfully synchronized {c} file(s)."
					error = f"Failed to synchronize {c} file(s)."
					command = Array(array=batch).string(joiner=" & ")
					response = self.utils.execute(command=command, error=error, message=msg, loader=loader)
					if not response.success:
						response = self.utils.execute(command=command, error=error, message=msg, loader=loader)
						if not response.success:
							return response

		# no multiprocessing.
		else:

			# sum together as one command
			sum = False

			# iterate updates.
			dirs, commands, c = {}, [], 0
			for path, info in updates.items():
				if self.log_level >= 0: print(f"Updating {path}.")

				# local to remote.
				if info["local_to_remote"]:
					if info["directory"] in [True, "True", "TRUE", "true"]: dirs[path] = info
					else: 
						commands += [self.local_to_remote(path, info, command=sum)]
						c += 1

				# remote to local.
				elif info["remote_to_local"]:
					if info["directory"] in [True, "True", "TRUE", "true"]: dirs[path] = info
					else: 
						commands += [self.remote_to_local(path, info, command=sum)]
						c += 1

			# iterate excepted dirs.
			for path, info in dirs.items():

				# local to remote.
				if info["local_to_remote"]:
					commands += [self.local_to_remote(path, info, directory=True, command=sum)]
					c += 1

				# remote to local.
				elif info["remote_to_local"]:
					commands += [self.remote_to_local(path, info, directory=True, command=sum)]
					c += 1

			# execute commands.
			if sum:
				loader = None
				if self.log_level >= 0: loader = f"Synchronizing {c} file(s)."
				msg = f"Successfully synchronized {c} file(s)."
				error = f"Failed to synchronize {c} file(s)."
				command = Array(array=commands).string(joiner=" && ")
				response = self.utils.execute(command=command, error=error, message=msg, loader=loader)
				if not response.success:
					response = self.utils.execute(command=command, error=error, message=msg, loader=loader)
					if not response.success:
						return response

		# push deletions.
		if multiprocessing:
			local_deletions, remote_deletions = [], []
			for path, info in deletions.items():
				if info["remote_to_local"]:
					local_deletions.append(path)
				else:
					remote_deletions.append(path)
			if local_deletions != []:
				if self.log_level >= 0: loader = dev0s.console.Loader(f"Deleting {len(local_deletions)} local file(s).")
				response = self.delete(local_deletions, remote=False)
				if self.log_level >= 0: loader.stop(success=response["success"])
				if not response["success"]: return response
			if remote_deletions != []:
				if self.log_level >= 0: loader = dev0s.console.Loader(f"Deleting {len(remote_deletions)} remote file(s).")
				response = self.delete(remote_deletions, remote=True)
				if self.log_level >= 0: loader.stop(success=response["success"])
				if not response["success"]: return response
			for path, info in deletions.items():
				self.reset_last_index(path, remote="both")
		else:
			for path, info in deletions.items():
				if info["remote_to_local"]:
					response = self.delete(path, remote=False)
				else:
					response = self.delete(path, remote=True)
				if not response["success"]: return response
				self.reset_last_index(path, remote="both")

		# check synchronized index.
		response = self.index(short=True)
		if not response["success"]: return response
		elif not response.synchronized:
			return dev0s.response.error(f"Failed to synchronize [{self.alias}:{self.remote}] & [{self.path}], mismatches: {response.mismatches}].")

		# handler.
		return dev0s.response.success(f"Successfully synchronized [{self.alias}:{self.path}] & [{self.path}].")
	


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
import os, sys
from dev0s.shortcuts import *

# index.
def index(path):
	indexed, dir, ids = Dictionary(path=False, dictionary={}), Files.Directory(path=path), []
	for _path_ in dir.paths(recursive=True, files_only=True, banned=[f"Icon\r"], banned_names=[".DS_Store", "._DS_Store", "__pycache__"]):
		if _path_ not in ids: 
			indexed[_path_] = gfp.mtime(path=_path_, format="seconds")
			ids.append(_path_)
	for _path_ in dir.paths(recursive=True, dirs_only=True, banned=[f"Icon\r"], banned_names=[".DS_Store", "._DS_Store", "__pycache__"]):
		id = _path_+" (d)"
		if os.listdir(_path_) == []: id += " (e)"
		if id not in ids:
			indexed[id] = gfp.mtime(path=_path_, format="seconds")
			ids.append(id)
	return indexed.sort(alphabetical=True)

# main.
if __name__ == "__main__":

	# arguments.
	path = dev0s.cli.get_argument("--path")
	json = dev0s.cli.arguments_present(["--json", "-j"])

	# checks.
	if not Files.exists(path):
		dev0s.response.log(response=dev0s.response.error(f"Path [{path}] does not exist."), json=json)
	elif not os.path.isdir(path):
		dev0s.response.log(response=dev0s.response.error(f"Path [{path}] is not a directory."), json=json)

	# handler.
	dict = index(path)
	dev0s.response.log(json=json, response=dev0s.response.success(f"Successfully indexed {len(dict)} files from directory [{path}].", {
		"index":dict,
	}))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from ssht00ls.classes.config import *
from ssht00ls.classes.aliases import aliases
from ssht00ls.classes.agent import agent
import ssht00ls.classes.ssync.utils as ssync_utils 
from ssht00ls.classes.ssync import daemons
from ssht00ls.classes.ssh import ssh
from ssht00ls.classes import utils

# the ssync object class.
class SSync(Traceback):
	def __init__(self,
		# initialize as specific not global (optional).
		#	the username.
		alias=None,
	):

		# docs.
		DOCS = {
			"module":"ssht00ls.ssync", 
			"initialized":True,
			"description":[], 
			"chapter": "Protocols", }

		# defaults.
		Traceback.__init__(self, traceback="ssht00ls.ssync", raw_traceback="ssht00ls.classes.ssync.SSync")

		# specific args.
		self.specific = alias != None
		self.alias = alias

		# defaults.
		self.utils = ssync_utils
		self.daemons = daemons
		
		#
	def index(self, path=None, alias=None, log_level=dev0s.defaults.options.log_level, checks=True, accept_new_host_keys=True):

		# check specific.
		if self.specific:
			if alias == None: alias = self.alias

		# checks.
		if path == None:
			return dev0s.response.error(f"Define parameter: path.")
		path = gfp.clean(path)

		# check encryption activated.
		if not ssht00ls_agent.activated:
			return dev0s.response.error(f"The {ssht00ls_agent.id} encryption requires to be activated.")

		# remote.
		if alias != None:

			# checks.
			if checks:
				
				# check alias.
				response = aliases.check(alias)
				if not response["success"]: return response

				# check passphrase.
				if CONFIG["aliases"][alias]["smartcard"] in [True, "true", "True"]:
					passphrase = CONFIG["aliases"][alias]["pin"]
				else:
					passphrase = CONFIG["aliases"][alias]["passphrase"]
				if passphrase not in ["", "none", "None", None]:
					response = ssht00ls_agent.encryption.decrypt()
					if not response["success"]: return response
					passphrase = response.decrypted.decode()
				else:
					passphrase = None
					
				# tests.
				response = agent.add(private_key=CONFIG["aliases"][alias]["private_key"], passphrase=passphrase)
				if not response["success"]: return response
				response = ssh.utils.test_ssht00ls(alias=alias, accept_new_host_keys=accept_new_host_keys)
				if not response["success"]: return response
				response = ssh.utils.test_path(alias=alias, path=path, accept_new_host_keys=accept_new_host_keys)
				if not response["success"]: return response

			# index.
			response = self.utils.execute(
				command=f"""printf 'yes' | ssh {DEFAULT_SSH_OPTIONS} {alias} ' export IPINFO_API_KEY="{IPINFO_API_KEY}" && python3 /usr/local/lib/ssht00ls/classes/ssync/index.py --path {path} --json --non-interactive --no-checks ' """,
				serialize=True,
				log_level=log_level,
				input={
					"Are you sure you want to continue connecting":Boolean(accept_new_host_keys).string(true="yes", false="no"),
				},
				optional=True,)
			if not response.success:
				return dev0s.response.error(f"Failed to connect with {alias}, error: {response.error}")
			else:
				return response

		# local.
		else:
			if checks:
				if not Files.exists(path):
					return dev0s.response.error(f"Path [{path}] does not exist.")
				elif not os.path.isdir(path):
					return dev0s.response.error(f"Path [{path}] is not a directory.")

			# handler.
			dict = self.utils.index(path)
			return dev0s.response.success(f"Successfully indexed {len(dict)} files from directory [{path}].", {
				"index":dict,
			})

			#
	def set_mounted_icon(self, path):
		if dev0s.defaults.vars.os in ["osx", "macos"]:
			#os.system(f"cp '{SOURCE_PATH}/static/media/icons/Icon\r' '{path}/Icon\r'")
			icon = f'{SOURCE_PATH}/static/media/icons/mounted.png'
			dev0s.utils.__execute_script__(f"""
				
				#!/bin/bash
				# Set Icon to a File / Folder on macOS

				icon="{icon}"
				dest="{path}"

				# Check inputs

				if [ ! -f $icon ]; then 
					echo "ERROR: File $1 does not exists"
					exit 1
				elif [[ ! $icon =~ .*\.(png|PNG|jpg|JPG) ]]; then
					echo "ERROR: Icon must be a .png|.jpg file"
					exit 1
				elif [ -f $dest ]; then
					folder=false
				elif [ -d $dest ]; then
					folder=true
				else
					echo 'ERROR: File|Folder destination does not exists'
					exit 1
				fi

				# Create icns icon

				sips -i $icon > /dev/null
				DeRez -only icns $icon > /tmp/tmpicns.rsrc

				# Set Icon

				if [ "$folder" = true ]; then
					Rez -append /tmp/tmpicns.rsrc -o $dest$'/Icon\r'
					SetFile -a C $dest
					SetFile -a V $dest$'/Icon\r'
				else
					Rez -append /tmp/tmpicns.rsrc -o $dest
					SetFile -a C $dest
				fi

				# Clean up

				rm /tmp/tmpicns.rsrc
				exit 0

				""")
	# pull & push.
	def pull(self,
		# the local path.
		path=None, 
		# the ssht00ls alias.
		alias=None, 
		# the remote path.
		remote=None, 
		# exlude subpaths (list) (leave None to exclude none).
		exclude=[],
		# path is directory boolean (leave None to parse automatically).
		directory=True,
		empty_directory=False,
		# update deleted files.
		delete=False,
		# forced mode.
		forced=False,
		# version control.
		safe=False,
		# accept new hosts keys.
		accept_new_host_keys=True,
		# checks.
		checks=True,
		# log level.
		log_level=dev0s.defaults.options.log_level,
		# get the command in str.
		command=False,
	):	

		# check specific.
		if self.specific:
			if alias == None: alias = self.alias

		# check encryption activated.
		if not ssht00ls_agent.activated:
			return dev0s.response.error(f"The {ssht00ls_agent.id} encryption requires to be activated.")
		if checks:
			if self.specific:
				try: self.activated
				except: self.activated = False
			if not self.specific or not self.activated:
				response = aliases.check(alias)
				if not response["success"]: return response
				response = agent.check(public_key=CONFIG["aliases"][alias]["public_key"])
				if not response.success:
					if "is not added to the" not in response.error: return response
					else:
						if CONFIG["aliases"][alias]["smartcard"] in [True, "true", "True"]:
							passphrase = CONFIG["aliases"][alias]["pin"]
						else:
							passphrase = CONFIG["aliases"][alias]["passphrase"]
						if passphrase not in [None, "None", "none", "", False]:
							response = ssht00ls_agent.encryption.decrypt(passphrase)
							if not response["success"]: return response
							passphrase = response.decrypted.decode()
						else:
							passphrase = None
						response = agent.add(private_key=CONFIG["aliases"][alias]["private_key"], passphrase=passphrase)
						if not response["success"]: return response
				if self.specific: self.activated = True
		return self.utils.pull(
			# the local path.
			path=path, 
			# the ssht00ls alias.
			alias=alias, 
			# the remote path.
			remote=remote, 
			# exlude subpaths (list) (leave None to use default).
			exclude=exclude,
			# path is directory boolean (leave None to parse automatically).
			directory=directory,
			empty_directory=empty_directory,
			# update deleted files.
			delete=delete,
			# forced mode.
			forced=forced,
			# version control.
			safe=safe,
			# accept new hosts keys.
			accept_new_host_keys=accept_new_host_keys,
			# checks.
			checks=checks,
			# log level.
			log_level=log_level,
			# get the command in str.
			command=command,)
	def push(self,
		# the local path.
		path=None, 
		# the ssht00ls alias.
		alias=None, 
		# the remote path.
		remote=None, 
		# exlude subpaths (list) (leave None to exclude none).
		exclude=[],
		# path is directory boolean (leave None to parse automatically).
		directory=True,
		empty_directory=False,
		# update deleted files.
		delete=False,
		# forced mode.
		forced=False,
		# version control.
		safe=False,
		# accept new hosts keys.
		accept_new_host_keys=True,
		# checks.
		checks=True,
		check_base=True,
		# log level.
		log_level=dev0s.defaults.options.log_level,
		# get the command in str.
		command=False,
	):

		# check specific.
		if self.specific:
			if alias == None: alias = self.alias

		# check encryption activated.
		if not ssht00ls_agent.activated:
			return dev0s.response.error(f"The {ssht00ls_agent.id} encryption requires to be activated.")
		if checks:
			if self.specific:
				try: self.activated
				except: self.activated = False
			if not self.specific or not self.activated:
				response = aliases.check(alias)
				if not response["success"]: return response
				response = agent.check(public_key=CONFIG["aliases"][alias]["public_key"])
				if not response.success:
					if "is not added to the" not in response.error: return response
					else:
						if CONFIG["aliases"][alias]["smartcard"] in [True, "true", "True"]:
							passphrase = CONFIG["aliases"][alias]["pin"]
						else:
							passphrase = CONFIG["aliases"][alias]["passphrase"]
						if passphrase not in [None, "None", "none", "", False]:
							response = ssht00ls_agent.encryption.decrypt(passphrase)
							if not response["success"]: return response
							passphrase = response.decrypted.decode()
						else:
							passphrase = None
						response = agent.add(private_key=CONFIG["aliases"][alias]["private_key"], passphrase=passphrase)
						if not response["success"]: return response
				if self.specific: self.activated = True
		return self.utils.push(
			# the local path.
			path=path, 
			# the ssht00ls alias.
			alias=alias, 
			# the remote path.
			remote=remote, 
			# exlude subpaths (list) (leave None to use default).
			exclude=exclude,
			# path is directory boolean (leave None to parse automatically).
			directory=directory,
			empty_directory=empty_directory,
			# update deleted files.
			delete=delete,
			# forced mode.
			forced=forced,
			# version control.
			safe=safe,
			# accept new hosts keys.
			accept_new_host_keys=accept_new_host_keys,
			# checks.
			checks=checks,
			check_base=check_base,
			# log level.
			log_level=log_level,
			# get the command in str.
			command=command,)
	# depricated.
	def mount(self, 
		# the local path.
		path=None, 
		# the remote path.
		remote=None, 
		# ssh alias.
		alias=None,
		# forced.
		forced=False,
		# exclude.
		exclude=['__pycache__', '.DS_Store'],
		# accept new host verification keys.
		accept_new_host_keys=True,
		# log level.
		log_level=dev0s.defaults.options.log_level,
		# the daemon mode.
		mode="mount",
	):

		# depricated.
		return dev0s.response.error("DEPRICATED.")
		
		# check specific.
		if self.specific:
			if alias == None: alias = self.alias

		# checks.
		response = dev0s.response.parameters.check(
			traceback=self.__traceback__(function="mount"), 
			parameters={
				"path:str":path,
				"remote:str":remote,
				"alias:str":alias,
				"exclude:list":exclude,
				"forced:bool":forced,
				"mode:str":mode,
				"accept_new_host_keys:bool":accept_new_host_keys, 
			})
		if not response["success"]: return response
		if mode in ["synchronize", "sync", "synch"]: mode = "sync"
		if mode not in ["mount", "sync"]:
			return dev0s.response.error(f"Specified an invalid mode: [{mode}], options: [mount, sync].")
		cache_path = gfp.absolute(path=gfp.clean(path.split(" (d)")[0], remove_last_slash=True))
		if mode in ["mount"] and not forced and Files.exists(path):
			return dev0s.response.error(f"Path [{path}] already exists.")
		elif self.daemons.running(cache_path):
			return dev0s.response.error(f"Path [{path}] is already mounted.")
		path = gfp.clean(path=path)
		remote = gfp.clean(path=remote)

		# pull.
		if mode in ["mount"]:
			response = self.pull(
				path=path, 
				alias=alias, 
				remote=remote, 
				exclude=exclude,
				forced=forced,
				delete=True,
				safe=False,
				log_level=log_level,)
			if not response["success"]: return response
		else:
			# make sure both dirs exist in the end.
			# raise error when both dirs do not exist at start.
			response = ssh.utils.test_dir(path=remote, alias=alias)
			remote_exists, local_exists = True, Files.exists(path)
			if not response["success"]: 
				if "does not exist" in response["error"]:
					remote_exists = False
				else:
					return response
			if not remote_exists and not local_exists:
				return dev0s.response.error(f"Both local and remote directories {path} & {alias}:{remote} do not exist.")
			if not local_exists:
				os.system(f"mkdir -p {path}")
				if not Files.exists(path): 
					return dev0s.response.error(f"Failed to create directory {path}.")
			if not remote_exists:
				response = ssh.utils.test_dir(path=remote, alias=alias, create=True)
				if not response["success"]: 
					return dev0s.response.error(f"Failed to create directory {alias}:{remote}, error: {response['error']}")

		# start daemon.
		cache.set(id=cache_path, group="daemons", data=f"*running* (timestamp={Date().seconds_timestamp})")
		self.set_mounted_icon(path)
		if not self.daemons.running(cache_path):
			return dev0s.response.error(f"Failed to set the {path} daemon status.")
		response =  self.daemon(alias=alias, path=path, remote=remote, start=True, mode=mode)
		if not response.success: return response

		# handler.
		if mode == "sync":
			return dev0s.response.success(f"Successfully synchronized {alias}:{remote} to {path}.")
		else:
			return dev0s.response.success(f"Successfully mounted {alias}:{remote} to {path}.")

		#
	def unmount(self, 
		# the local path.
		path=None, 
		# forced required.
		forced=False,
		# sudo required.
		sudo=False,
		# log level.
		log_level=dev0s.defaults.options.log_level,
	):

		# depricated.
		return dev0s.response.error("DEPRICATED.")

		# checks.
		response = dev0s.response.parameters.check(
			traceback=self.__traceback__(function="unmount"), 
			parameters={
				"path:str":path,
				"forced:bool":forced,
				"sudo:bool":sudo, 
			})
		if not response["success"]: return response
		path = gfp.clean(path)
		cache_path = gfp.absolute(gfp.clean(path.split(" (d)")[0], remove_last_slash=True))
		if not self.daemons.running(cache_path):
			if not Files.exists(path):
				return dev0s.response.error(f"Path [{path}] does not exist.")
			elif not os.path.isdir(path):
				return dev0s.response.error(f"Path [{path}] is not a directory.")
			status = str(cache.get(id=cache_path, group="daemons"))
			if not self.daemons.running(cache_path):
				return dev0s.response.error(f"Path [{path}] is not mounted (status: {status}).")

		# wait for daemon stop.
		response = self.daemons.stop(path=path)
		if not success: return response

		# handler.
		return dev0s.response.success(f"Successfully unmounted [{path}].")

		#
	def daemon(self, 
		# the ssh alias.
		alias=None, 
		# the remote path.
		remote=None, 
		# thel local path.
		path=None, 
		# settings.
		start=True,
		# the daemon mode.
		mode="mount",
		# the daemons log level.
		log_level=dev0s.defaults.log_level(default=-1),
		# sandbox (do not delete any files).
		sandbox=False,
		# overwrite sleeptime.
		sleeptime=SSYNC_DAEMON_SLEEPTIME,
		# serialized.
		serialized={},
	):
		# depricated.
		return dev0s.response.error("DEPRICATED.")

		# check specific.
		if self.specific:
			if alias == None: alias = self.alias

		# serialzied.
		if serialized != {}:
			alias, remote, path, start, mode, log_level, sandbox, sleeptime = Dictionary(path=False, dictionary=serialized).unpack({
				"alias":None,
				"remote":None,
				"path":None,
				"start":True,
				"mode":"mount",
				"log_level":dev0s.defaults.log_level(default=-1),
				"sandbox":False,
				"sleeptime":SSYNC_DAEMON_SLEEPTIME,
			})
		if mode in ["synchronize", "sync", "synch"]: mode = "sync"
		if mode not in ["mount", "sync"]:
			return dev0s.response.error(f"Specified an invalid mode: [{mode}], options: [mount, sync].")
		path = gfp.clean(path)
		remote = gfp.clean(remote)
		_daemon_ = daemons.Daemon({
			"alias":alias,
			"remote":remote,
			"path":path,
			"mode":mode,
			"log_level":log_level,
			"sandbox":sandbox,
			"ssync":self,
			"utils":self.utils,
			"sleeptime":sleeptime,
		})
		if start: ssht00ls_agent.webserver.start_thread(_daemon_, group="daemons", id=_daemon_.id)
		return dev0s.response.success("Successfully initialized the daemon", {
			"daemon":_daemon_,
		})
	
# initialized objects.
ssync = SSync()
if CHECKS and dev0s.defaults.options.interactive and not dev0s.cli.argument_present("--reset-cache"):
	daemons.sync()
	


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
# keep this import style so you can also use the keys.Keys() etc directly.
from ssht00ls.classes.config import ssht00ls_agent
from ssht00ls.classes.config import NETWORK_INFO as network_info

# import classes.
from ssht00ls.classes import utils, installation, sshd, sshfs, scp, agent, keys, connections, smartcards, aliases, ssh, ssync, client, smb



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from ssht00ls.classes.config import *
import ssht00ls.classes.ssh.utils as ssh_utils 

# the ssh tunnel object class.
class Tunnel(Thread):
	def __init__(self,
		# initialize as specific not global (optional).
		# 	the alias.
		alias=None,
		# 	the tunnel ip.
		ip=None,
		# 	the local port.
		port=None,
		# 	the remote port.
		remote_port=None,
		# 	the reconnect boolean.
		reconnect=False,
		# 	the thread's sleeptime.
		sleeptime=60,
		# 	the reconnect reattemps.
		reattemps=15,
		# 	the log level.
		log_level=0,
	):


		# docs.
		DOCS = {
			"module":"ssht00ls.ssh.Tunnel", 
			"initialized":False,
			"description":[], 
			"chapter": "Connections", }

		# defaults.
		Thread.__init__(self, traceback="ssht00ls.ssh.tunnel", raw_traceback="ssht00ls.ssh.Tunnel", log_level=dev0s.defaults.log_level(default=-1))

		# modules.
		self.utils = ssh_utils

		# specific args.
		self.specific = alias != None and ip != None and port != None and remote_port != None
		self.alias = alias
		self.ip = ip
		self.port = port
		self.remote_port = remote_port
		self.reconnect = reconnect
		self.log_level = log_level
		self.sleeptime = sleeptime
		self.reattemps = reattemps

		#
	# functions.
	def establish(self,
		# the alias.
		alias=None,
		# the tunnel ip.
		ip=None,
		# the local port.
		port=None,
		# the remote port.
		remote_port=None,
		# the reconnect boolean (leave None to use False).
		reconnect=None,
		# the log level (leave None to use 0).
		log_level=None,
	):
		
		# check specific.
		if alias == None: alias = self.alias
		if ip == None: ip = self.ip
		if port == None: port = self.port
		if remote_port == None: remote_port = self.remote_port
		if reconnect == None: reconnect = self.reconnect
		if log_level == None: log_level = self.log_level # keep this one indent back.

		# check parameters.
		response = dev0s.response.parameters.check(
			traceback=self.__traceback__(function="establish"),
			parameters={
				"alias":alias,
				"ip":ip,
				"port":port,
				"remote_port":remote_port,
				"reconnect":reconnect, 
			})
		if not response["success"]: return response

		# id.
		id = self.__id__(alias=alias, ip=ip, port=port, remote_port=remote_port)

		# check already established
		if self.established:
			return dev0s.response.success(f"Tunnel [{id}] is already established.")

		# execute.
		output = dev0s.code.execute(f"ssh -L {port}:{ip}:{remote_port} -f -N {DEFAULT_SSH_OPTIONS} {alias}", async_=True)
		if not output.success: return output
		if output != "":
			return dev0s.response.error(f"Failed to establish tunnel {id}, error: {output}.")

		# send cached run permission.
		response = ssht00ls_agent.webserver.set(group="tunnels.run_permission", id=id, data="True")
		if not response.success: return response

		# tunnel.
		if not self.specific:
			tunnel = Tunnel(
				alias=alias,
				ip=ip,
				port=port,
				remote_port=remote_port,
				reconnect=reconnect,)
			if not tunnel.established:
				return dev0s.response.error(f"Failed to establish tunnel {tunnel.id}.")
		elif not self.established:
			return dev0s.response.error(f"Failed to establish tunnel {self.id}.")

		# start thread.
		if not self.specific and reconnect:
			#response = tunnel.start()
			response = ssht00ls_agent.webserver.start_thread(tunnel, group="tunnels", id=tunnel.id)
			if not response.success: return response

		# handler.
		attributes = {}
		if not self.specific: attributes["tunnel"] = tunnel
		return dev0s.response.success(f"Successfully established tunnel [{id}].", attributes)

		#
	def kill(self,
		# the alias.
		alias=None,
		# the tunnel ip.
		ip=None,
		# the local port.
		port=None,
		# the remote port.
		remote_port=None,
		# the log level.
		log_level=None,
	):

		# check specific.
		if alias == None: alias = self.alias
		if ip == None: ip = self.ip
		if port == None: port = self.port
		if remote_port == None: remote_port = self.remote_port
		if log_level == None: log_level = self.log_level

		# check parameters.
		response = dev0s.response.parameters.check(
			traceback=self.__traceback__(function="kill"),
			parameters={
				"alias":alias,
				"port":port, 
			})
		if not response["success"]: return response

		# id.
		id = self.__id__(alias=alias, ip=ip, port=port, remote_port=remote_port)

		# check established.
		if not self.__established__(alias=alias, ip=ip, port=port, remote_port=remote_port):
			return dev0s.response.success(f"Tunnel [{id}] was not active.")

		# check stop thread.
		response = ssht00ls_agent.webserver.set(group="tunnels.run_permission", id=id, data="False")
		if not response.success: return response

		"""
		tunnel = None
		response = ssht00ls_agent.webserver.get_thread(group="tunnels", id=id)
		if not response.success and "There is no thread cached for" not in response.error:
			return response
		elif response.success:
			tunnel = response.thread
		if tunnel != None:
			response = tunnel.stop()
			if not response.success: return response
		"""

		# kill pid(s).
		pid = self.__pid__(alias=alias, ip=ip, port=port, remote_port=remote_port)
		if pid == None:
			return dev0s.response.error(f"Unable to determine the pid of tunnel [{id}].")
		elif isinstance(pid, (list, Array)):
			for i in pid:
				response = dev0s.code.kill(pid=i, log_level=log_level)
				if not response.success:
					return dev0s.response.error(f"Failed to kill pid [{i}] of tunnel [{id}], error: {response.error}")
		else:
			response = dev0s.code.kill(pid=pid, log_level=log_level)
			if not response.success:
				return dev0s.response.error(f"Failed to kill pid [{pid}] of tunnel [{id}], error: {response.error}")


		# check stopped.
		if self.__established__(alias=alias, ip=ip, port=port, remote_port=remote_port):
			return dev0s.response.error(f"Failed to stop tunnel [{id}].")

		# handler.
		return dev0s.response.success(f"Successfully stopped tunnel [{id}].")

		#
	def list(self, alias=None):
		if alias == None: alias = self.alias
		array, dictionary = [], {}
		#all_aliases, c = list(CONFIG["aliases"].keys()), 0
		includes = f"ssh -L "
		response = dev0s.code.processes(includes=includes)
		if not response.success: return response
		for pid, info in response.processes.items():
			pass_ = False
			alias_ = None
			try:
				port,ip,remote_port = info["command"].split("ssh -L ")[1].split(" ")[0].split(":")
			except:
				pass_ = True
			try:
				alias_ = info["command"].split(" -f -N ")[1].split(" ")[0]
			except:
				pass_ = True
			if not pass_ and (alias in [None, "*", "all"] or alias == alias_):
				id = self.__id__(alias=alias_, ip=ip, port=port, remote_port=remote_port)
				array.append(id)
				dictionary[id] = {
					"alias":alias_,
					"ip":ip,
					"port":port,
					"remote_port":remote_port,
				}
		return dev0s.response.success(f"Successfully listed {len(array)} tunnels.", {
			"tunnels":array,
			"array":array,
			"dictionary":dictionary,
		})
	def iterate(self, alias=None):
		if alias == None: alias = self.alias
		return self.list(alias=alias)["dictionary"].items()
		#
	# thread.
	def __run__(self):

		# checks.
		if not self.established:
			self.send_crash(response=dev0s.response.error_response("The tunnel is not established yet."))

		# start.
		while self.run_permission:

			# check no cached run permission.
			response = ssht00ls_agent.webserver.get(group="tunnels.run_permission", id=id)
			if not response.success: 
				self.send_crash(response=response)
				break
			elif not bool(response.data):
				break

			# check no longer mounted.
			if not os.path.exists(self.path):
				break

			# check if connection is still active otherwise reconnect.
			if not self.established:
				crashed = False
				for attempt in range(self.reattemps):
					response = self.establish()
					if response.success:
						break
					else:
						if attempt >= self.reattemps-1:
							self.send_crash(response=response)
							crashed = True
							break
						else:
							time.sleep(self.sleeptime)
				if crashed: break
					
			# sleep.
			time.sleep(self.sleeptime) 

		#
	def __stop__(self):
		response = self.kill()
		if not response.success: return response
		return dev0s.response.success(f"Successfully stopped [{self.id}]")
	# properties.
	@property
	def id(self):
		return self.__id__()
	def __id__(self, alias=None, ip=None, port=None, remote_port=None):
		if alias == None: alias = self.alias
		if port == None: port = self.port
		if remote_port == None: remote_port = self.remote_port
		if ip == None: ip = self.ip
		return f"{port}:{ip}:{remote_port}:{alias}"
	@property
	def established(self):
		return self.__established__()
	def __established__(self, alias=None, ip=None, port=None, remote_port=None):
		if alias == None: alias = self.alias
		if port == None: port = self.port
		if remote_port == None: remote_port = self.remote_port
		if ip == None: ip = self.ip
		id = self.__id__(alias=alias, ip=ip, port=port, remote_port=remote_port)
		response = dev0s.code.processes(includes=f"ssh -L {port}:{ip}:{remote_port}")
		if not response.success: raise ValueError(f"Unable to determine if tunnel {id} is active, error: {response.error}")
		return len(response.processes) >= 1
	@property
	def pid(self):
		return self.__pid__()
	def __pid__(self, alias=None, ip=None, port=None, remote_port=None):
		if alias == None: alias = self.alias
		if port == None: port = self.port
		if remote_port == None: remote_port = self.remote_port
		if ip == None: ip = self.ip
		id = self.__id__(alias=alias, ip=ip, port=port, remote_port=remote_port)
		response = dev0s.code.processes(includes=f"ssh -L {port}:{ip}:{remote_port}")
		if not response.success: raise ValueError(f"Unable to determine if tunnel {id} is active, error: {response.error}")
		pids = list(response.processes.keys())
		if len(pids) == 0:
			return None
		elif len(pids) == 1:
			return pids[0]
		else:
			return pids
	# bool representation.
	def __bool__(self):
		return bool(self.established)
	# str representation.
	def __str__(self):
		if self.specific:
			return f"{self.traceback[:-1]} ({self.id}) (established: {self.established}) (pid: {self.pid}) >"
		else:
			return self.traceback
	# representation.
	def __repr__(self):
		return str(self)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# alias.
ALIAS = "ssht00ls"

# updates.
import os, sys
if "--update" in sys.argv and ALIAS in sys.argv[0]:
	os.system(f"curl -s https://raw.githubusercontent.com/vandenberghinc/{ALIAS}/master/{ALIAS}/requirements/installer.remote | bash ")
	sys.exit(0)

# imports.
import os, sys, subprocess, pexpect, random, getpass, time

# inc imports.
from dev0s.shortcuts import *

# source.	
SOURCE_PATH = dev0s.defaults.source_path(__file__, back=3)
BASE = dev0s.defaults.source_path(SOURCE_PATH)
dev0s.defaults.operating_system(supported=["linux", "macos"])
dev0s.defaults.alias(alias=ALIAS, executable=f"{SOURCE_PATH}")
if dev0s.defaults.options.log_level >= 2:
	dev0s.response.log(f"{ALIAS}:")
	dev0s.response.log(f"  * source: {SOURCE_PATH}")

# universal options.
# interactive must be False by default.
CHECKS = not dev0s.cli.arguments_present(["--no-checks"])
RESET_CACHE = dev0s.cli.arguments_present("--reset-cache")
if dev0s.defaults.options.log_level >= 2:
	dev0s.response.log(f"  * interactive: {dev0s.defaults.options.interactive}")
	dev0s.response.log(f"  * checks: {CHECKS}")

# database.
DATABASE = Directory(path=dev0s.env.get_string("SSHT00LS_DATABASE", default=f"{dev0s.defaults.vars.home}/.{ALIAS}"))
if not DATABASE.fp.exists():
	dev0s.response.log(f"{color.orange}Root permission{color.end} required to create {ALIAS} database [{DATABASE}].")
	os.system(f" sudo mkdir -p {DATABASE}")
	Files.chown(str(DATABASE), owner=dev0s.defaults.vars.owner, group=dev0s.defaults.vars.group, sudo=True, recursive=True)
	Files.chmod(str(DATABASE), permission=700, sudo=True, recursive=True)

# config.
CONFIG = Dictionary(path=dev0s.env.get_string("SSHT00LS_CONFIG", default=DATABASE.join("config","")), load=True, default={})

# logs.
if dev0s.defaults.options.log_level >= 2:
	dev0s.response.log(f"  * database: {DATABASE}")
	dev0s.response.log(f"  * config: {CONFIG.fp}")

# initialize cache.
cache = dev0s.database.Database(
	path=gfp.clean(f"{DATABASE}/.cache/"))

# dev0s settings.
IPINFO_API_KEY = os.environ.get("IPINFO_API_KEY")

# ssh settings.
SSH_TIMEOUT = int(dev0s.cli.get_argument("--timeout", required=False, default=10))
SSH_REATTEMPS = int(dev0s.cli.get_argument("--reattempts", required=False, default=3))
DEFAULT_SSH_OPTIONS = f"-o ConnectTimeout={SSH_TIMEOUT} -o ConnectionAttempts={SSH_REATTEMPS}"

# daemon settings.
SSYNC_DAEMON_SLEEPTIME = round(float(dev0s.cli.get_argument("--daemon-sleeptime", required=False, default=0.25)), 2)

# logs.
if dev0s.defaults.options.log_level >= 2:
	dev0s.response.log(f"  * ssh timeout: {SSH_TIMEOUT}")
	dev0s.response.log(f"  * ssh reattempts: {SSH_REATTEMPS}")
	dev0s.response.log(f"  * daemon sleeptime: {SSYNC_DAEMON_SLEEPTIME}")

# speed up non interactive.
if CHECKS and not RESET_CACHE:

	# network.
	NETWORK_INFO = dev0s.network.info()
	if not NETWORK_INFO["success"]: 
		dev0s.response.log(error=NETWORK_INFO.error, json=dev0s.cli.arguments_present(["--json", "-j"]), log_level=0)
		sys.exit(1)
	if dev0s.defaults.options.log_level >= 2:
		dev0s.response.log(f"  * public ip: {NETWORK_INFO['public_ip']}")
		dev0s.response.log(f"  * private ip: {NETWORK_INFO['private_ip']}")

	# check lib.
	if not Files.exists(f"{SOURCE_PATH}/lib") or dev0s.cli.argument_present("--download-lib"):
		dev0s.response.log("Downloading the ssht00ls library.")
		os.system(f"rm -fr /tmp/ssht00ls && git clone -q https://github.com/vandenberghinc/ssht00ls /tmp/ssht00ls && rsync -azq /tmp/ssht00ls/ssht00ls/lib/ {Files.join(SOURCE_PATH, 'lib/')}")
		if dev0s.cli.argument_present("--download-lib"): sys.exit(0)

	# check usr lib.
	if not Files.exists("/usr/local/lib/ssht00ls"):
		dev0s.response.log(f"{color.orange}Root permission{color.end} required to install the ssht00ls system library.")
		os.system(f" sudo rsync -azq --delete {SOURCE_PATH}/ /usr/local/lib/ssht00ls")
		Files.chown("/usr/local/lib/ssht00ls", owner=dev0s.defaults.vars.owner, group=dev0s.defaults.vars.group, sudo=True, recursive=True)
		Files.chmod("/usr/local/lib/ssht00ls", permission=770, sudo=True, recursive=True)

	# database.
	for dir, permission in [
		[f"{dev0s.defaults.vars.home}/.{ALIAS}", 770],
		[f"{dev0s.defaults.vars.home}/.{ALIAS}/lib", 770],
		[f"{dev0s.defaults.vars.home}/.{ALIAS}/.cache", 770],
	]:
		if not Files.exists(dir): 
			os.system(f"sudo mkdir {dir} && sudo chown {dev0s.defaults.vars.user}:{dev0s.defaults.vars.group} {dir} && sudo chmod {permission} {dir}")

	# files.
	CONFIG.check(save=True, default={
		"aliases":{
			"example.com (key's are optional)":{
				"username":"administrator",
				"public_ip":"192.168.1.100",
				"public_port":22,
				"private_ip":"84.84.123.192",
				"private_port":22,
				"private_key":"~/keys/example.com/administrator/private_key",
				"public_key":"~/keys/example.com/administrator/public_key",
				"passphrase":None,
				"smartcard":False,
				"pin":None,
			}
		},
	})

	# database.
	Files.chmod(f"{SOURCE_PATH}/lib/utils/*", permission="+x")

	# agent.
	ssht00ls_agent = dev0s.encryption.Agent(
		# id.
		id="ssht00ls-agent",
		# cache.
		database=DATABASE.join(".agent/"),
		# webserver.
		host="127.0.0.1",
		port=2379,
		# encryption.
		private_key=dev0s.env.get("SSHT00LS_PRIVATE_KEY", default=DATABASE.join("keys/master/private_key")),
		public_key=dev0s.env.get("SSHT00LS_PUBLIC_KEY", default=DATABASE.join("keys/master/public_key")),
		passphrase=None, 
		interactive=dev0s.defaults.options.interactive,
		# traceback.
		traceback="ssht00ls.ssht00ls_agent",
	)

	# webserver.
	if dev0s.cli.argument_present("--stop-agent"):
		response = ssht00ls_agent.webserver.stop()
		if response.success:
			dev0s.response.log(response=response, json=dev0s.defaults.options.json)
			sys.exit(0)
		else:
			dev0s.response.log(response=response, json=dev0s.defaults.options.json)
			sys.exit(1)
	elif dev0s.defaults.options.interactive and not ssht00ls_agent.webserver.running: # is also automatically done in agent.generate & agent.activate
		if dev0s.defaults.options.log_level >= 2:
			dev0s.response.log(f"  * forking the ssht00ls agent.")
		response = ssht00ls_agent.webserver.fork()
		dev0s.response.log(response=response, json=dev0s.defaults.options.json)
		if not response.success: sys.exit(1)
	if dev0s.defaults.options.log_level >= 2:
		dev0s.response.log(f"  * webserver: {ssht00ls_agent.webserver}")

	# check interactive.
	if dev0s.defaults.options.interactive:

		# generate encryption.
		if not ssht00ls_agent.generated:
			response = ssht00ls_agent.generate()
			dev0s.response.log(response=response, json=dev0s.defaults.options.json)
			if not response.success: sys.exit(1)

		# activate encryption.
		else:
			response = ssht00ls_agent.activate()
			dev0s.response.log(response=response, json=dev0s.defaults.options.json)
			if not response.success: sys.exit(1)

#



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from ssht00ls.classes.config import * 
from ssht00ls.classes import utils
import os, sys, json, subprocess, pexpect

# check default errors..
def check_errors(output):
	for i in [
		"\nrsync: ",
		"\nrsync error: ",
		"\nssh: ",
		"\nssh error: ",
		"\nsshfs: ",
		"\nsshfs error: ",
		"\nscp: ",
		"\nscp error: ",
		"\nmount_smbfs: ",
		"\nmount_smbfs error: ",
		"\nclient_loop: ",
	]:
		if i in "\n"+output:
			e = str(String(("\n"+output).split(i)[1].split("\n")[0]).capitalized_word())
			while True:
				if len(e) > 0 and e[len(e)-1] in [" ", ".", "\n"]: e = e[:-1]
				elif len(e) > 0 and e[0] in [" "]: e = e[1:]
				else: break
			return dev0s.response.error(e+" ("+i.replace("\n","").replace(" ","").replace(":","").replace(" error","")+").")
	return dev0s.response.success("The output contains no (default) errors.")
	"""
	for i in [
		"rsync: ", "rsync error: ",
		"ssh: ", "ssh error: ",
		"sshfs: ", "sshfs error: ",
		"scp: ", "scp error: ",
		"mount_smbfs: ", "mount_smbfs error: ",
		"client_loop: send disconnect: Broken pipe",
	]:
		if i in output:
			e = String(i+output.split(i)[1].split("\n")[0]).capitalized_word().replace("Client_loop: send disconnect:", "Client loop: send disconnect:").replace("Ssh", "SSH")
			while True:
				if len(e) > 0 and e[len(e)-1] in [" ", "."]: e = e[:-1]
				elif len(e) > 0 and e[0] in [" "]: e = e[1:]
				else: break
			return dev0s.response.error(e+".")
	return dev0s.response.success("The output contains no (default) errors.")
	"""

# execute command.
def execute( 
	# 
	# Command:
	#   the command in str.
	command="ssh <alias> ' ls ' ",
	#
	# Options:
	#   asynchronous process.
	async_=False,
	#	await asynchronous child (sync process always awaits).
	wait=False,
	#	kill process when finished (async that is not awaited is never killed).
	kill=True,
	#   the subprocess shell parameter.
	shell=False,
	#   serialize output to dict (expect literal dictionary / json output).
	serialize=False,
	# accept new host keys.
	accept_new_host_keys=True,
	#
	# Input (sync only):
	#   send input to the command.
	#	  undefined: send no input & automatically await the process since input is always sync.
	#	  dict instance: selects "and" mode ; send expected inputs and their value & return error when one of them is missing.
	#	  list[dict] instance: send all dictionaries in the list (default dict behaviour so one of the keys in each dict is expected).
	input=None,
	#   the input timeout (float) (list with floats by index from input)
	timeout=2.0,
	#   do not throw an error when the input is missing or not expected when optional is disabled (bool).
	optional=False, 
	#	apped default accept host keys input.
	append_default_input=True,
	#
	# Logging.
	# the success message (leave None to use the default).
	message=None,
	#   loader message (str, Loader).
	loader=None,
	# stop the loader at the end of the request (bool).
	stop_loader=True,
	#   the log level (int).
	log_level=dev0s.defaults.options.log_level,
	#
	# System functions.
	#   add additional attributes to the spawn object.
	__spawn_attributes__={},
):

	# execute.
	if log_level >= 6: print(command)
	if message != None: message = message.replace("$COMMAND", command)

	# default input.
	default_input = {
		"Are you sure you want to continue connecting":Boolean(accept_new_host_keys).string(true="yes", false="no"),
	}
	if input.__class__.__name__ in ["list", "Array"]:
		if append_default_input:
			input = [default_input] + input
	elif input.__class__.__name__ in ["dict", "Dictionary"]:
		if append_default_input:
			input = [
				default_input,
				input,
			]
	elif append_default_input:
		input = default_input
		optional = True

	# execute.
	final_response = dev0s.code.execute(
		command=command,
		input=input,
		optional=optional,
		async_=async_,
		wait=wait,
		kill=kill,
		shell=shell,
		serialize=serialize,
		loader=loader,
		stop_loader=stop_loader,
		log_level=log_level,
		__spawn_attributes__=__spawn_attributes__,	)
	if not final_response.success: return final_response
	if message != None: final_response.message = message

	# check errors.
	response = check_errors(final_response.output)
	if not response.success: return response
	return final_response




















	#########################
	#
	# DEPRICATED
	# old.
	# script.

	# version 1.
	#output = dev0s.utils.__execute_script__(command)
	
	# version 2.
	#try:
	#	output = subprocess.check_output(["sh", path]).decode()
	#except subprocess.CalledProcessError as e:
	#	return dev0s.response.error(f"Failed to execute command [{command}], (output: {e.output}), (error: {e}).")

	# version 3.
	#response = dev0s.code.execute(
	#	command=command,)
	#if not response["success"]: return response
	#output = response.output
	# equal to:
	path = f"/tmp/tmp_script_{String('').generate()}"
	Files.save(path, command)
	try:
		proc = subprocess.run(
		    ["sh", path],
		    stdin=subprocess.PIPE,
			check=True,
			capture_output=True,
			text=True,
		)
	except subprocess.CalledProcessError as error:
		error_, output = error.stderr, error.output
		if isinstance(error_, bytes): error_ = error_.decode()
		if isinstance(output, bytes): output = output.decode()
		if loader != None: loader.stop(success=False)
		if log_level <= 0:
			return dev0s.response.error(f"Failed to execute command ({command}), (error: {error_}).")
		else:
			return dev0s.response.error(f"Failed to execute command ({command}), (error: {error_}), (output: {output}).")
	error_, output = proc.stderr, proc.stdout
	if isinstance(error_, bytes): error_ = error_.decode()
	if isinstance(output, bytes): output = output.decode()
	if error_ != "":
		if loader != None: loader.stop(success=False)
		if log_level <= 0:
			return dev0s.response.error(f"Failed to execute command ({command}), (error: {error_}).")
		else:
			return dev0s.response.error(f"Failed to execute command ({command}), (error: {error_}), (output: {output}).")
	if len(output) > 0 and output[len(output)-1] == "\n": output = output[:-1]
	Files.delete(path)

	# handler.
	response = check_errors(output)
	if not response.success:
		if loader != None: loader.stop(success=False)
		return response
		#print(output)
		#return dev0s.response.error(error)
	else:
		if serialize:
			try: response = dev0s.response.ResponseObject(output)
			except Exception as e: 
				if loader != None: loader.stop(success=False)
				return dev0s.response.error(f"Failed to serialize output: {output}")
			if loader != None: loader.stop()
			return response
		else:
			if loader != None: loader.stop()
			if get_output:
				return dev0s.response.success(message, {
					"output":output,
				})
			else:
				return dev0s.response.success(message)

	#

# test ssh functions.
def test(alias=None, accept_new_host_keys=True, checks=True):

	# init.
	command = f"""ssh {DEFAULT_SSH_OPTIONS} {alias} ' echo "Hello World" ' """
	
	# version 3.
	response = dev0s.code.execute(
		command=command,
		input={
			"Are you sure you want to continue connecting":Boolean(accept_new_host_keys).string(true="yes", false="no"),
		},
		optional=True,)
	if not response.success:
		return dev0s.response.error(f"Failed to connect with {alias}, error: {response.error}")
	output = response.output
	response = check_errors(output)
	if "Hello World" in output:
		return dev0s.response.success(f"Successfully connected with {alias}.")
	else:
		return dev0s.response.error(f"Failed to connect with {alias}, error: {output}")

	"""
	version 2.
	# pexpect.
	spawn = dev0s.console.Spawn(command)
	response = spawn.start()
	if not response.success: return response

	# expect.
	response = spawn.expect(timeout=1.0, data=[
		"Are you sure you want to continue connecting",
	])
	if not response.success:
		if "None of the specified inputs were expected." not in response.error:
			return response
		else:
			a=1 # skip not required.

	# send.
	elif response.success:
		if response.index == 0:
			response = spawn.send(timeout=0.5, data={
				"Are you sure you want to continue connecting":"yes",
			})
			if not response.success: return response
		else: raise exceptions.InvalidUsage(f"Missed expected spawn index: [{response.index}].")

	# handler.
	response = spawn.output()
	if not response.success: return response
	output = response.output
	response = check_errors(output)
	if not response.success:
		return dev0s.response.error(f"Failed to connect with {alias}, error: {output}")
	elif "Hello World" in output:
		return dev0s.response.success(f"Successfully connected with {alias}.")
	else:
		return dev0s.response.error(f"Failed to connect with {alias}, error: {output}")
	"""

	"""
	# version 1
	if dev0s.defaults.options.log_level >= 6:
		print(f"<{ALIAS}.ssh.utils.test> command: {command}")
	response = dev0s.code.execute(command)
	if not response.success: return dev0s.response.error(response.error)
	output = response.output
	response = check_errors(output)
	if not response.success:
		return dev0s.response.error(f"Failed to connect with {alias}, error: {output}")
	elif "Hello World" in output:
		return dev0s.response.success(f"Successfully connected with {alias}.")
	else:
		return dev0s.response.error(f"Failed to connect with {alias}, error: {output}")
	"""
def test_path(alias=None, path=None, accept_new_host_keys=True, checks=True):
	if checks:
		response = test(alias=alias, accept_new_host_keys=accept_new_host_keys)
		if not response.success: return response
	command = f"""ssh {DEFAULT_SSH_OPTIONS} {alias} ' ls -ld {path} ' """
	if dev0s.defaults.options.log_level >= 6:
		print(f"<{ALIAS}.ssh.utils.test> command: {command}")
	response = dev0s.code.execute(
		command=command,
		input={
			"Are you sure you want to continue connecting":Boolean(accept_new_host_keys).string(true="yes", false="no"),
		},
		optional=True,)
	if not response.success:
		return dev0s.response.error(f"Failed to connect with {alias}, error: {response.error}")
	output = response.output
	response = check_errors(output)
	if not response.success:
		return dev0s.response.error(f"Path {alias}:{path} does not exist.")
	elif "No such file or directory" not in output:
		return dev0s.response.success(f"Path {alias}:{path} exists.")
	else:
		return dev0s.response.error(f"Path {alias}:{path} does not exist.")
def test_dir(alias=None, path=None, accept_new_host_keys=True, create=False, created=False, checks=True):
	if checks:
		response = test(alias=alias, accept_new_host_keys=accept_new_host_keys)
		if not response.success: return response
	lpath = "\'"+path+"\'"
	command = f"""ssh {DEFAULT_SSH_OPTIONS} {alias} ''' .ssht00ls/utils/isdir {path}''' """
	if dev0s.defaults.options.log_level >= 6:
		print(f"<{ALIAS}.ssh.utils.test> command: {command}")
	response = dev0s.code.execute(
		command=command,
		input={
			"Are you sure you want to continue connecting":Boolean(accept_new_host_keys).string(true="yes", false="no"),
		},
		optional=True,)
	if not response.success:
		return dev0s.response.error(f"Failed to connect with {alias}, error: {response.error}")
	output = response.output
	if output.replace("\n","") == "directory":
		return dev0s.response.success(f"Path {alias}:{path} is a directory.", {
			"created":created,
		})
	elif output.replace("\n","") in ["directory", "does-not-exist"]:
		if output.replace("\n","") in ["does-not-exist"]:
			if create:
				output = dev0s.utils.__execute_script__(f"""ssh {DEFAULT_SSH_OPTIONS} {alias} ''' mkdir -p {path}''' """)
				if "permission denied" in output: return dev0s.response.error(f"Unable to create remote directory [{alias}:{path}].")
				return test_dir(alias=alias, path=path, checks=False, create=False, created=True)
			else:
				return dev0s.response.error(f"Path {alias}:{path} does not exist.")
		else:
			return dev0s.response.error(f"Path {alias}:{path} is not a directory.")
	else:
		return dev0s.response.error(f"Unable to check remote directory {alias}:{path}, output: {output}.")
def test_ssht00ls(alias=None, accept_new_host_keys=True, install=True):
	for path in [f"/usr/local/lib/{ALIAS}"]:
		response = test_path(alias=alias, accept_new_host_keys=accept_new_host_keys, path=path)
		if not response.success:
			if response.error == f"Path {alias}:{path} does not exist.":
				if install:
					loader = dev0s.console.Loader(f"Installing ssht00ls library on remote {alias}.")
					response = dev0s.code.execute(f"ssh {DEFAULT_SSH_OPTIONS} {alias} ' curl https://raw.githubusercontent.com/vandenberghinc/{ALIAS}/master/{ALIAS}/requirements/installer.remote | bash ' ")
					if not response.success: return dev0s.response.error(response.error)
					output = response.output
					response = test_ssht00ls(alias=alias, accept_new_host_keys=accept_new_host_keys, install=False)
					loader.stop(success=response.success)
					#print(output)
					return response
				else:
					return dev0s.response.error(f"Remote {alias} does not have library ssht00ls installed.")
			else:
				return response
	return dev0s.response.success(f"Remote {alias} has library ssht00ls installed.")

# check / start the ssh agent.
def ssh_agent():
	return utils.ssh_agent()

# kill all ssh procs with that includes the identifier.
def kill(identifier=None, sudo=False):
	response = dev0s.response.parameters.check({
		"identifier:str":identifier,})
	if not response.success: return response
	return dev0s.code.kill(includes=identifier, sudo=sudo)
	# old.
	killed = 0
	output = dev0s.utils.__execute_script__(f"""ps -ax | grep "{identifier}" | """ + """awk '{print $1"|"$2"|"$3"|"$4}' """)
	for line in output.split("\n"):
		if line not in ["", " "]:
			pid,tty,_,process = line.split("|")
			if process not in dont_kill:
				loader = dev0s.console.Loader(f"Killing process {pid}.")
				if sudo: _sudo_ = "sudo "
				else: _sudo_ = ""
				output = dev0s.utils.__execute_script__(f"{_sudo_}kill {pid}")
				if "terminated" in output:
					loader.stop()
					killed += 1
				else:
					loader.stop(success=False)
					dev0s.response.log(f"Failed to stop process {pid}, error: {output}")
	return dev0s.response.success(f"Successfully killed {killed} process(es) that included identifier [{identifier}].")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from ssht00ls.classes.config import *
from ssht00ls.classes import utils
from ssht00ls.classes.ssh import ssh, Tunnel
from ssht00ls.classes.aliases import aliases

# the smb client object class.
class SMB(Thread):
	def __init__(self,
		# initialize as specific not global (optional).
		# 	the share id (#1).
		id=None,
		# 	the mountpoint path (#2).
		path=None,
		# 	the alias (#3).
		alias=None,
		# 	the server's ip (leave None to retrieve from alias).
		ip=None,
		# 	the server's port.
		port=445,
		# 	tunnel smb through ssh.
		tunnel=False,
		tunnel_obj=None, # do not use the tunnel_obj parameter.
		# 	the reconnect boolean (only used whe tunnel is enabled).
		reconnect=False,
		# 	the thread's sleeptime.
		sleeptime=60,
		# 	the reconnect reattemps.
		reattemps=15,
		# 	the log level.
		log_level=dev0s.defaults.options.log_level,
	):


		# docs.
		DOCS = {
			"module":"ssht00ls.smb", 
			"initialized":True,
			"description":[], 
			"chapter": "Protocols", }

		# defaults.
		Thread.__init__(self, traceback="ssht00ls.smb", log_level=dev0s.defaults.log_level(default=-1))

		# specific variables.
		self.specific = alias != None and id != None and path != None and ip != None
		self.id_ = id
		self.path = path
		self.alias = alias
		self.ip = ip
		self.port = port
		self.tunnel = tunnel
		self.tunnel_obj = tunnel_obj
		self.reconnect = reconnect
		self.sleeptime = sleeptime
		self.reattemps = reattemps
		self.log_level = log_level

		#
	# functions.
	def mount(self,
		# the share id (leave None to use self.id) (required) (#1).
		id=None,
		# the mountpoint path (leave None to use self.path) (required) (#2).
		path=None,
		# the ssh alias (leave None to use self.alias) (required) (#3).
		alias=None,
		# the username of for the remote server (leave None to use the aliases username).
		username=None,
		# the password of the remote server's user (optional) (leave None to prompt) (use "" for no passphrase).
		password=None,
		# the ip of the remote server (leave None to use the aliases ip).
		ip=None,
		# the port (leave None to use self.port).
		port=None,
		# tunnel over ssh (leave None to use self.tunnel).
		tunnel=None,
		# the reconnect boolean (only used whe tunnel is enabled) (leave None to use self.reconnect).
		reconnect=None,
		# the log level (leave None to use self.log_level).
		log_level=None,
	):
		

		# check specific.
		if id == None: id = self.id_
		if path == None: path = self.path
		if alias == None: alias = self.alias
		if ip == None: ip = self.ip
		if tunnel == None: tunnel = self.tunnel
		if reconnect == None: reconnect = self.reconnect
		if port == None: port = self.port 
		if log_level == None: log_level = self.log_level 

		# loader.
		if log_level >= 0:
			loader = dev0s.console.Loader(f"Mounting smb share [{alias}:{id}] to [{path}]")

		# checks.
		response = dev0s.response.parameters.check({
			"id:str,String":id,
			"path:str,String":path,
			"alias:str,String":alias,
			"port:int,Integer":port,
			"tunnel:bool,Boolean":tunnel,
			"reconnect:bool,Boolean":reconnect,
		}, traceback=self.__traceback__(function="mount"))
		if not response.success: 
			if log_level >= 0: loader.stop(success=False)
			return response

		# check alias.
		response = aliases.info(alias=alias)
		if not response.success: 
			if log_level >= 0: loader.stop(success=False)
			return response
		if username == None: username = response.info["username"]
		if ip == None: ip = response.info["public_ip"]

		# parse smb.
		response = self.parse(path=path)
		if not response.success: 
			if log_level >= 0: loader.stop(success=False)
			return response
		elif response.mounted:
			if log_level >= 0: loader.stop(success=False)
			return dev0s.response.error(f"Path [{path}] is already mounted.")
		elif not response.exists:
			os.system(f"mkdir -p {path} 2> /dev/null && chown {dev0s.defaults.vars.user}:{dev0s.defaults.vars.group} {path}")
			if not os.path.exists(path):
				os.system(f"sudo mkdir -p {path} 2> /dev/null && sudo chown {dev0s.defaults.vars.user}:{dev0s.defaults.vars.group} {path}")

		# check tunnel.
		if tunnel:
			new_tunnel = False
			if not self.specific or (self.specific and self.tunnel_obj == None):
				# check already established tunnel from previous session.
				established_port = None
				response = dev0s.code.processes(includes=f":localhost:445 -f -N {DEFAULT_SSH_OPTIONS} {alias}")
				if not response.success: 
					if log_level >= 0: loader.stop(success=False)
					return response
				elif len(response.processes) > 0:
					command = Dictionary(response.processes)[0]["command"]
					try: 
						s = str(command).split(f":localhost:445 -f -N {DEFAULT_SSH_OPTIONS} {alias}")[0].split(" ")
						established_port = int(s[len(s)-1])
					except: a=1
				# create new tunnel.
				remote_port = int(port)
				if established_port == None:
					response = dev0s.network.free_port(start=6000)
					if not response.success: 
						if log_level >= 0: loader.stop(success=False)
						return response
					port = response.port
				else:
					port = established_port
				ip = "localhost"
				tunnel_ = Tunnel(
					alias=alias,
					ip=ip,
					port=port,
					remote_port=remote_port,
					reconnect=reconnect,)
				if self.specific:
					self.tunnel_obj = tunnel_
				new_tunnel = True
			elif self.specific and self.tunnel_obj != None:
				tunnel_ = self.tunnel_obj
			if not tunnel_.established:
				response = tunnel_.establish()
				if not response.success: 
					if log_level >= 0: loader.stop(success=False)
					return success
				time.sleep(0.5)
				if reconnect and new_tunnel:
					#tunnel_.start()
					response = ssht00ls_agent.webserver.start_thread(tunnel_, group="tunnels", id=tunnel_.id)
					if not response.success: 
						if log_level >= 0: loader.stop(success=False)
						return response

		# command.
		if dev0s.defaults.vars.os in ["macos"]:
			if password == None:
				user_pass = username
			elif password == "":
				user_pass = f"{username}:''"
			else:
				user_pass = f"{username}:{password}"
			command = f"mount_smbfs //{user_pass}@{ip}:{port}/{id} {path}"
		else:
			if log_level >= 0: loader.stop(success=False)
			return dev0s.response.error("Coming soon.")
			command = ""

		# execute.
		try:
			output = dev0s.code.execute(command)
		except KeyboardInterrupt as e:
			if dev0s.defaults.vars.os in ["macos"]:
				dev0s.code.kill(includes=command, log_level=-1)
			raise KeyboardInterrupt(e)
		if not output.success: 
			if log_level >= 0: loader.stop(success=False)
			return output

		# reconnect smb thread.
		if reconnect:
			if self.specific:
				smb = self
			else:
				smb = SMB(
					id=id,
					path=path,
					alias=alias,
					ip=ip,
					port=port,
					tunnel=tunnel,
					tunnel_obj=tunnel_,
					reconnect=reconnect,)
			if reconnect:
				#smb.start()
				response = ssht00ls_agent.webserver.start_thread(smb, group="smb.mounts", id=path)
				if not response.success: 
					if log_level >= 0: loader.stop(success=False)
					return response

		# handlers.
		attributes = {}
		if not self.specific and tunnel: attributes["tunnel"] = tunnel_
		if not self.specific and reconnect: attributes["smb"] = smb
		if log_level >= 0: loader.stop()
		return dev0s.response.success(f"Successfully mounted smb share [{alias}:{id}] to [{path}].", attributes)

		#
	def unmount(self, 
		# the mountpoint path (leave None to use self.path) (required) (#1).
		path=None, 
		# the forced umount option.
		forced=False, 
		# root permission required for force.
		sudo=False,
		# the log level (leave None to use self.log_level).
		log_level=None,
	):

		# specific.
		if path == None: path = self.path
		if log_level == None: log_level = self.log_level 

		# loader.
		if log_level >= 0:
			loader = dev0s.console.Loader(f"Unmounting [{path}]")

		# checks.
		response = dev0s.response.parameters.check({
			"path:str,String":path,
			"forced:bool,Boolean":forced,
			"sudo:bool,Boolean":sudo,
		}, traceback=self.__traceback__(function="unmount"))
		if not response["success"]: 
			if log_level >= 0: loader.stop(success=False)
			return response

		# parse smb.
		response = self.parse(path=path)
		if not response.success: 
			if log_level >= 0: loader.stop(success=False)
			return response
		elif not response.mounted:
			if log_level >= 0: loader.stop(success=False)
			return dev0s.response.error(f"Path [{path}] is not mounted.")

		# check stop thread.
		"""
		smb = None
		response = ssht00ls_agent.webserver.get_thread(group="smb.mounts", id=path)
		print(response)
		if not response.success and "There is no thread cached for" not in response.error:
			return response
		elif response.success:
			smb = response.thread
		if smb != None:
			response = smb.stop()
			dev0s.response.log(response=response)
			if not response.success: return response
		"""

		# exceute.
		command = ""
		if sudo: 
			command += "sudo "
		command += "umount "
		if forced: 
			command += "-f "
		command += path
		output = dev0s.code.execute(command)

		# handlers.
		if not output:
			if log_level >= 0: loader.stop(success=False)
			return dev0s.response.error(f"Failed to unmount [{path}], error: {output}")
		if output != "":
			if log_level >= 0: loader.stop(success=False)
			return dev0s.response.error((f"Failed to unmount [{path}], error: "+output.replace("\n", ". ").replace(". .", ".")+".)").replace(". .",".").replace("\r","").replace("..","."))
		else:
			if log_level >= 0: loader.stop()
			return dev0s.response.success(f"Successfully unmounted [{path}].")

		#
	def parse(self,
		# the mountpoint path (leave None to use self.path) (required) (#1). 
		path=None,
	):

		# specific.
		if path == None: path = self.path

		# checks.
		response = dev0s.response.parameters.check({
			"path:str,String":path,
		}, traceback=self.__traceback__(function="unmount"))
		if not response["success"]: return response

		# handlers.
		try: mounted = os.path.ismount(path)
		except FileNotFoundError: mounted = False
		try: directory = os.path.isdir(path)
		except FileNotFoundError: directory = False
		return dev0s.response.success(f"Successfully parsed [{path}].", {
			"exists":os.path.exists(path),
			"mounted":mounted,
			"directory":directory,
		})

		#
	# thread.
	def __run__(self):

		# checks.
		if not self.mounted:
			self.send_crash(response=dev0s.response.error_response("The smb share is not mounted yet."))

		# start.
		while self.run_permission:

			# check no longer mounted.
			if not os.path.exists(self.path):
				break

			# check if connection is still active otherwise reconnect.
			if not self.mounted:
				crashed = False
				for attempt in range(self.reattemps):
					response = self.mount()
					if response.success:
						break
					else:
						if attempt >= self.reattemps-1:
							self.send_crash(response=response)
							crashed = True
							break
						else:
							time.sleep(self.sleeptime)
				if crashed: break
					
			# sleep.
			time.sleep(self.sleeptime)

		#
	def __stop__(self):
		# does never needs to be called since the thread auto stops.
		# do not kill the tunnel since it may be used by other mounts.
		response = self.unmount()
		if not response.success and "is not mounted" not in response.error: return response
		return dev0s.response.success(f"Successfully stopped [{self.id}]")
	# properties.
	@property
	def id(self):
		return self.__id__()
	def __id__(self, alias=None, id=None, path=None):
		if alias == None: alias = self.alias
		if id == None: id = self.id_
		if path == None: path = self.path
		return f"{alias}:{id} {path}"
	@property
	def mounted(self):
		return self.__mounted__()
	def __mounted__(self, alias=None, id=None, path=None):
		if alias == None: alias = self.alias
		if id == None: id = self.id_
		if path == None: path = self.path
		id = self.__id__(alias=alias, id=id, path=path)
		if not os.path.exists(path): return False
		return os.path.ismount(path)
	# bool representation.
	def __bool__(self):
		return bool(self.mounted)
	# str representation.
	def __str__(self):
		if self.specific:
			return f"{self.traceback[:-1]} ({self.id}) (mounted: {self.mounted}) >"
		else:
			return self.traceback
	# representation.
	def __repr__(self):
		return str(self)

	#

# initialized objects.
smb = SMB()

# testing.
"""
response = smb.mount(
	id="packages",
	path="/Volumes/packages",
	alias="dev.vandenberghinc.com",
	password="",
	tunnel=True,
	reconnect=True,)
print("response:",response)
time.sleep(5)
response = smb.unmount(
	path="/Volumes/packages",)
print(response)
quit()
quit()
"""
#
	